unit Dxsect;

{-------------------------------------------------------------------}
{                    Unit:    Dxsect.pas                            }
{                    Project: EPA SWMM                              }
{                    Version: 5.0                                   }
{                    Date:    3/31/05    (5.0.005)                  }
{                             6/19/07    (5.0.010)                  }
{                             10/27/08   (5.0.014)                  }
{                    Author:  L. Rossman                            }
{                                                                   }
{   Dialog form unit used to edit cross-sectional geometry for      }
{   a conduit.                                                      }
{-------------------------------------------------------------------}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, NumEdit, ImgList, Spin, Buttons, ComCtrls,
  Uproject, Uglobals;

type
  TXsectShape = record
    Text : array[1..4] of String;
  end;

const
  MAX_SHAPE_INDEX = 24;                                                         //(5.0.010 - LR)
  CUSTOM_SHAPE_INDEX = 23;                                                      //(5.0.010 - LR)
  IRREG_SHAPE_INDEX = 22;
  HORIZ_ELLIPSE_SHAPE_INDEX = 7;
  VERT_ELLIPSE_SHAPE_INDEX = 8;
  ARCH_SHAPE_INDEX = 9;
  TRAPEZOIDAL_SHAPE_INDEX = 5;
  CIRC_SHAPE_INDEX = 0;
  FORCE_MAIN_INDEX = 1;
  MODBASKET_SHAPE_INDEX = 14;                                                   //(5.0.014 - lr)

  // This is the list of cross-sectional shapes that are available
  XsectShapes: array[0..MAX_SHAPE_INDEX] of TXsectShape =
    ((Text: ('CIRCULAR', '', '', '')),
     (Text: ('FORCE_MAIN', 'Roughness*', '', '')),
     (Text: ('FILLED_CIRCULAR', 'Filled Depth', '', '')),
     (Text: ('RECT_CLOSED', 'Bottom Width', '', '')),
     (Text: ('RECT_OPEN', 'Bottom Width', '', '')),
     (Text: ('TRAPEZOIDAL', 'Bottom Width', 'Left Slope', 'Right Slope')),
     (Text: ('TRIANGULAR', 'Top Width', '', '')),
     (Text: ('HORIZ_ELLIPSE', 'Max. Width', '', '')),
     (Text: ('VERT_ELLIPSE', 'Max. Width', '', '')),
     (Text: ('ARCH', 'Max. Width', '', '')),
     (Text: ('PARABOLIC', 'Top Width', '', '')),
     (Text: ('POWER', 'Top Width', 'Exponent', '')),
     (Text: ('RECT_TRIANGULAR', 'Top Width', 'Triangle Height', '')),
     (Text: ('RECT_ROUND', 'Top Width', 'Bottom Radius', '')),
     (Text: ('MODBASKETHANDLE', 'Bottom Width', 'Top Radius', '')),             //(5.0.014 - LR)
     (Text: ('EGG', '', '', '')),
     (Text: ('HORSESHOE', '', '', '')),
     (Text: ('GOTHIC', '', '', '')),
     (Text: ('CATENARY', '', '', '')),
     (Text: ('SEMIELLIPTICAL', '', '', '')),
     (Text: ('BASKETHANDLE', '', '', '')),
     (Text: ('SEMICIRCULAR', '', '', '')),
     (Text: ('IRREGULAR', '', '', '')),
     (Text: ('CUSTOM', '', '', '')),
     (Text: ('DUMMY', '', '', ''))                                              //(5.0.010 - LR)
     );

  XsectNotes: array[0..MAX_SHAPE_INDEX] of String =
  ('Standard circular pipe.',
   'Circular pipe with alternate friction loss equation for pressurized flow.',
   'Circular pipe partly filled with sediment.',
   'Closed rectangular box conduit.',
   'Open rectangular channel.',
   'Open trapezoidal channel. Slopes are horizontal / vertical.',
   'Open triangular channel.',
   'Closed horizontal elliptical pipe. For standard shapes, enter 0 for width.',
   'Closed vertical elliptical pipe. For standard shapes, enter 0 for width.',
   'Closed arch pipe. For std. shapes, enter code # for depth and 0 for width.',
   'Open parabolic channel with sidewall shape'#10'y = C x^2.',
   'Open channel with sidewall shape'#10'y = C x^n.',
   'Closed rectangular top with triangular bottom.',
   'Closed rectangular top with circular bottom.',
   'Rectangular bottom with closed circular top.',
   'Classic Phillips Standard Egg sewer shape.',
   'Classic Boston Horseshoe sewer shape.',
   'Classic Gothic sewer shape.',
   'Classic Catenary sewer shape (i.e., inverted egg).',
   'Classic Louisville Semielliptic sewer shape.',
   'Classic Baskethandle sewer shape.',
   'Classic Semicircular sewer shape.',
   'Open irregular natural channel described by transect coordinates.',
   'Closed custom shape described by a user-supplied shape curve.',
   'There are no parameters associated with a dummy cross-section.');

type
  TXsectionForm = class(TForm)
    OKBtn: TButton;
    CancelBtn: TButton;
    HelpBtn: TButton;
    Label1: TLabel;
    NumEdit1: TNumEdit;
    Label2: TLabel;
    NumEdit2: TNumEdit;
    Label3: TLabel;
    NumEdit3: TNumEdit;
    Label4: TLabel;
    NumEdit4: TNumEdit;
    TsectCombo: TComboBox;
    ShapeListBox: TComboBox;
    ShapeLabel: TLabel;
    BarrelsLabel: TLabel;
    TsectBtn: TBitBtn;
    BarrelsEdit: TEdit;
    BarrelsUpDown: TUpDown;
    Image1: TImage;
    TsectLabel: TLabel;
    Bevel1: TBevel;
    Bevel2: TBevel;
    SpecialNote: TLabel;
    DimenLabel: TLabel;
    UnitsLabel: TEdit;
    ForceMainNote: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure NumEdit1Change(Sender: TObject);
    procedure HelpBtnClick(Sender: TObject);
    procedure OKBtnClick(Sender: TObject);
    procedure ShapeListBoxClick(Sender: TObject);
    procedure TsectBtnClick(Sender: TObject);
  private
    { Private declarations }
    OldShapeIndex: Integer;
    function  ValidateData: Boolean;
    function  ValidateTextEntry(NumEdit: TNumEdit; const N: Integer;
              const I: Integer): Boolean;
  public
    { Public declarations }
    HasChanged: Boolean;
    procedure SetData(const Shape: String; const Geom1: String;
              const Geom2: String; const Geom3: String; const Geom4: String;
              const Barrels: String; const Tsect: String);
    procedure GetData(var Shape: String; var Geom1: String;
              var Geom2: String; var Geom3: String; var Geom4: String;
              var Barrels: String; var Tsect: String);
  end;

//var
//  XsectionForm: TXsectionForm;

implementation

{$R *.DFM}
{$R Xsects.res}

uses
  Uedit;

const
  MSG_BAD_DATA = 'Invalid data. Please re-enter a value.';
//  TXT_ARCH_NOTE =
//  'For standard arch shapes, enter code number for depth and 0 for width.';
//  TXT_ELLIPSE_NOTE = 'For standard ellipse shapes, enter 0 for width.';
  TXT_TRANSECT_NOTE = 'Transect Name';
  TXT_CUSTOM_NOTE = 'Shape Curve Name';
  TXT_FEET = 'Feet';
  TXT_METERS = 'Meters';
  TXT_MAX_DEPTH = 'Max. Depth';
  TXT_INCHES = ' (in)';
  TXT_MMETERS = ' (mm)';
  TXT_C_FACTOR = '*Hazen-Williams C-factor';
  TXT_ROUGHNESS = '*Darcy-Weisbach roughness height';

procedure TXsectionForm.FormCreate(Sender: TObject);
//-----------------------------------------------------------------------------
// OnCreate handler for form.
//-----------------------------------------------------------------------------
begin
  // Set font size
  Uglobals.SetFont(self);

  // Load names of project's transects into transect combo box
  TsectCombo.Items := Project.Lists[TRANSECT];
  SpecialNote.Caption := '';

  // Position controls on the form
  TsectLabel.Left := NumEdit1.Left;
  TsectLabel.Top := ShapeLabel.Top;
  TsectLabel.Visible := False;

  TsectCombo.Left := TsectLabel.Left;
  TsectCombo.Top := ShapeListBox.Top;
  TsectCombo.Visible := False;

  TsectBtn.Top := TsectCombo.Top;
  TsectBtn.Left := TsectCombo.Left + TsectCombo.Width;
  TsectBtn.Visible := False;

  ForceMainNote.Left := Label1.Left;
  ForceMainNote.Top := SpecialNote.Top;
  ForceMainNote.Width := SpecialNote.Width;
  ForceMainNote.Visible := False;

  // Assign text to units label
  with UnitsLabel do
  begin
    if Uglobals.UnitSystem = usUS
    then Text := TXT_FEET
    else Text := TXT_METERS;
  end;

  if SameText(Project.Options.Data[FORCE_MAIN_EQN_INDEX], 'H-W') then
    ForceMainNote.Caption := TXT_C_FACTOR
  else if Uglobals.UnitSystem = usUS
  then ForceMainNote.Caption := TXT_ROUGHNESS + TXT_INCHES
  else ForceMainNote.Caption := TXT_ROUGHNESS + TXT_MMETERS;
end;

procedure TXsectionForm.ShapeListBoxClick(Sender: TObject);
//-----------------------------------------------------------------------------
// OnClick handler for the cross-section shape selection list box.
// Customizes form depending on what shape is chosen.
//-----------------------------------------------------------------------------
var
  I: Integer;
  S: String;
begin
    I := ShapeListBox.ItemIndex;

    // Load proper shape image into the Image control
    if I < MAX_SHAPE_INDEX then
    begin
      S := 'XSECT' + IntToStr(I);
      Image1.Picture.Bitmap.Handle := LoadBitmap(hInstance, PChar(S));
      Image1.Visible := True;
    end
    else Image1.Visible := False;
    SpecialNote.Caption := XsectNotes[I];

    // Asssign visibility & labels to the dimension edit controls
    TsectLabel.Visible := False;
    TsectCombo.Visible := False;
    TsectBtn.Visible := False;
//    SpecialNote.Visible := False;
//    SlopeLabel.Visible := False;
    Label1.Caption := TXT_MAX_DEPTH;
    Label2.Caption := XsectShapes[I].Text[2];
    Label3.Caption := XsectShapes[I].Text[3];
    Label4.Caption := XsectShapes[I].Text[4];
    NumEdit1.Visible := True;
    NumEdit2.Visible := (Length(XsectShapes[I].Text[2]) > 0);
    NumEdit3.Visible := (Length(XsectShapes[I].Text[3]) > 0);
    NumEdit4.Visible := (Length(XsectShapes[I].Text[4]) > 0);
    ForceMainNote.Visible := False;

    // Special case for Irregular cross sections
    if I = IRREG_SHAPE_INDEX then
    begin
      NumEdit1.Visible := False;
      TsectCombo.Items := Project.Lists[TRANSECT];                              //(5.0.010 - LR)
      TsectLabel.Top := ShapeLabel.Top;                                         //(5.0.010 - LR)
      TsectCombo.Top := ShapeListBox.Top;                                       //(5.0.010 - LR)
      TsectBtn.Top := TsectCombo.Top;                                           //(5.0.010 - LR)
      TsectLabel.Caption := TXT_TRANSECT_NOTE;                                  //(5.0.010 - LR)
      TsectLabel.Visible := True;
      TsectCombo.Visible := True;
      TsectBtn.Visible := True;
    end;

    // Special case for Custom shape                                            //(5.0.010 - LR)
    if I = CUSTOM_SHAPE_INDEX then                                              //(5.0.010 - LR)
    begin                                                                       //(5.0.010 - LR)
      TsectCombo.Items := Project.Lists[SHAPECURVE];                            //(5.0.010 - LR)
      TsectLabel.Top := Label3.Top;                                             //(5.0.010 - LR)
      TsectCombo.Top := NumEdit3.Top;                                           //(5.0.010 - LR)
      TsectBtn.Top := TsectCombo.Top;                                           //(5.0.010 - LR)
      TsectLabel.Caption := TXT_CUSTOM_NOTE;                                    //(5.0.010 - LR)
      TsectLabel.Visible := True;                                               //(5.0.010 - LR)
      TsectCombo.Visible := True;                                               //(5.0.010 - LR)
      TsectBtn.Visible := True;                                                 //(5.0.010 - LR)
    end;

    // Special case for Force Main shpae
    if I = FORCE_MAIN_INDEX then
    begin
      ForceMainNote.Visible := True;
    end;

    // Special case for Trapezoidal shape
//    if I = TRAPEZOIDAL_SHAPE_INDEX
//    then SlopeLabel.Visible := True;

    // Special case for Arch shape
//    if I = ARCH_SHAPE_INDEX then
//    begin
//      SpecialNote.Caption := TXT_ARCH_NOTE;
//      SpecialNote.Visible := True;
//      SpecialNote.Refresh;
//    end;

    // Special case for Elliptical shapes
//    if I in [HORIZ_ELLIPSE_SHAPE_INDEX, VERT_ELLIPSE_SHAPE_INDEX] then
//    begin
//      SpecialNote.Caption := TXT_ELLIPSE_NOTE;
//      SpecialNote.Visible := True;
//      SpecialNote.Refresh;
//    end;

    if I = MAX_SHAPE_INDEX then
    begin
      Label1.Caption := '';
      NumEdit1.Visible := False;
    end;

    Label1.Visible := NumEdit1.Visible;
    UnitsLabel.Visible := NumEdit1.Visible and BarrelsEdit.Enabled;
    DimenLabel.Visible := UnitsLabel.Visible;
    BarrelsLabel.Visible := NumEdit1.Visible;
    BarrelsEdit.Visible := NumEdit1.Visible;
    BarrelsUpDown.Visible := NumEdit1.Visible;
end;

procedure TXsectionForm.SetData(const Shape: String; const Geom1: String;
  const Geom2: String; const Geom3: String; const Geom4: String;
  const Barrels: String; const Tsect: String);
//-----------------------------------------------------------------------------
//  Loads a given set of cross-section parameters into the form.
//  Shape = name of cross section shape,
//  Geom1 .. Geom4 = shape's dimensional parameters
//  Barrels = number of barrels
//  Tsect = name of transect object for irregular shapes
//-----------------------------------------------------------------------------
var
  I         : Integer;
  K         : Integer;
  MaxShapes : Integer;
  S         : String;
begin
  // If Barrels string empty, then form is being used to set defaults
  if Length(Barrels) = 0 then
  begin
    BarrelsEdit.Enabled := False;
    BarrelsUpDown.Enabled := False;
    MaxShapes := IRREG_SHAPE_INDEX - 1;
  end
  else
  begin
    MaxShapes := MAX_SHAPE_INDEX;
    try
      I := StrToInt(Barrels);
    except
      On EConvertError do I := 1;
    end;
    if I >= 1 then BarrelsUpDown.Position := I;
  end;

  // Add remaining parameters to form
  S := Shape;
  TsectCombo.Text := Tsect;
  NumEdit1.Text := Geom1;
  NumEdit2.Text := Geom2;
  NumEdit3.Text := Geom3;
  NumEdit4.Text := Geom4;

  // Add names of xsection shapes to dropdown shape list box
  for I := 0 to MaxShapes do
  begin
    ShapeListBox.Items.Add(XsectShapes[I].Text[1]);
  end;

  // Find index of current shape in dropdown shape list box
  I := -1;
  for K := 0 to MaxShapes do
  begin
    if SameText(S, XsectShapes[K].Text[1]) then
    begin
      I := K;
      break;
    end;
  end;
  ShapeListBox.ItemIndex := I;
  ShapeListBoxClick(self);
  HasChanged := False;
  OldShapeIndex := I;
end;

procedure TXsectionForm.GetData(var Shape: String; var Geom1: String;
  var Geom2: String; var Geom3: String; var Geom4: String;
  var Barrels: String; var Tsect: String);
//-----------------------------------------------------------------------------
//  Retrieves cross-section parameters from the form.
//-----------------------------------------------------------------------------
var
  I: Integer;
begin
  with ShapeListBox do Shape := Items[ItemIndex];
  if NumEdit2.Visible then Geom2 := NumEdit2.Text else Geom2 := '0';            //(5.0.010 - LR)
  if NumEdit3.Visible then Geom3 := NumEdit3.Text else Geom3 := '0';            //(5.0.010 - LR)
  if NumEdit4.Visible then Geom4 := NumEdit4.Text else Geom4 := '0';            //(5.0.010 - LR)
  if ShapeListBox.ItemIndex = IRREG_SHAPE_INDEX then
  begin
    Geom1 := 'N/A';
    Barrels := '1';
    Tsect := TsectCombo.Text;
    I := Project.Lists[TRANSECT].IndexOf(Tsect);
    if I >= 0
    then Geom1 := TTransect(Project.Lists[TRANSECT].Objects[I]).Data[TRANSECT_MAX_DEPTH];
  end
  else if ShapeListBox.ItemIndex = CUSTOM_SHAPE_INDEX then                      //(5.0.010 - LR)
  begin                                                                         //(5.0.010 - LR)
    Geom1 := NumEdit1.Text;                                                     //(5.0.010 - LR)
    Geom2 := TsectCombo.Text;                                                   //(5.0.010 - LR)
    Barrels := IntToStr(BarrelsUpDown.Position);                                //(5.0.010 - LR)
    Tsect := TsectCombo.Text;                                                   //(5.0.010 - LR)
  end                                                                           //(5.0.010 - LR)
  else if ShapeListBox.ItemIndex = MAX_SHAPE_INDEX then
  begin
    Geom1 := '0';
    Barrels := '1';
    Tsect := '';
  end
  else
  begin
    Geom1 := NumEdit1.Text;
    Barrels := IntToStr(BarrelsUpDown.Position);
    Tsect := '';
  end;
  if ShapeListBox.ItemIndex <> OldShapeIndex then HasChanged := True;
end;

procedure TXsectionForm.NumEdit1Change(Sender: TObject);
//-----------------------------------------------------------------------------
// OnChange handler for all 4 NumEdit controls as well as the TsectCombo
// control.
//-----------------------------------------------------------------------------
begin
  HasChanged := True;
end;

procedure TXsectionForm.OKBtnClick(Sender: TObject);
//-----------------------------------------------------------------------------
// OnClick handler for the OK button.
//-----------------------------------------------------------------------------
begin
  if ValidateData then ModalResult := mrOK;
end;

function  TXsectionForm.ValidateData: Boolean;
//-----------------------------------------------------------------------------
// Validates data entries on the form.
//-----------------------------------------------------------------------------
var
  I: Integer;
begin
  Result := True;
  I := ShapeListBox.ItemIndex;
  if NumEdit1.Visible then Result := Result and ValidateTextEntry(NumEdit1, 1, I);
  if NumEdit2.Visible then Result := Result and ValidateTextEntry(NumEdit2, 2, I);
  if NumEdit3.Visible then Result := Result and ValidateTextEntry(NumEdit3, 3, I);
  if NumEdit4.Visible then Result := Result and ValidateTextEntry(NumEdit4, 4, I);
end;

function  TXsectionForm.ValidateTextEntry(NumEdit: TNumEdit; const N: Integer;
  const I: Integer): Boolean;
//-----------------------------------------------------------------------------
// Validates the entry made in a NumEdit control for a given shape.
//-----------------------------------------------------------------------------
var
  X: Single;
begin
  // Assume entry is 0 if edit control is blank
  Result := True;
  if Length(NumEdit.Text) = 0 then NumEdit.Text := '0';

  // Convert the string to a number
  X := StrToFloat(Trim(NumEdit.Text));

  // If the number is 0, then check if the type of shape permits it
  if (X = 0) then
  begin
    if (N = 2)
    and (I in [CIRC_SHAPE_INDEX, TRAPEZOIDAL_SHAPE_INDEX,
               HORIZ_ELLIPSE_SHAPE_INDEX, VERT_ELLIPSE_SHAPE_INDEX,
               ARCH_SHAPE_INDEX])
    then Exit;

    if (N = 3) and (I = MODBASKET_SHAPE_INDEX) then Exit;                       //(5.0.014 - LR)

    // Otherwise issue an error message
    MessageDlg(MSG_BAD_DATA, mtError, [mbOK], 0);
    if NumEdit.Visible then NumEdit.SetFocus;
    Result := False;
  end;
end;

procedure TXsectionForm.TsectBtnClick(Sender: TObject);
//-----------------------------------------------------------------------------
// OnClick handler for the TSectBtn control. Launches the Transect Editor
// dialog for the transect named in the TsectCombo box.
//-----------------------------------------------------------------------------
var
  I: Integer;
  S: String;
begin
  S := Trim(TsectCombo.Text);

  if ShapeListBox.ItemIndex = CUSTOM_SHAPE_INDEX then                           //(5.0.010 - LR)
  begin                                                                         //(5.0.010 - LR)
    I := Project.Lists[SHAPECURVE].IndexOf(S);                                  //(5.0.010 - LR)
    S := Uedit.EditCurve(SHAPECURVE, I);                                        //(5.0.010 - LR)
    if Length(S) > 0 then                                                       //(5.0.010 - LR)
    begin                                                                       //(5.0.010 - LR)
      TsectCombo.Text := S;                                                     //(5.0.010 - LR)
      TsectCombo.Items := Project.Lists[SHAPECURVE];                            //(5.0.010 - LR)
    end;                                                                        //(5.0.010 - LR)
    Exit;                                                                       //(5.0.010 - LR)
  end;                                                                          //(5.0.010 - LR)

  I := Project.Lists[TRANSECT].IndexOf(S);
  S := Uedit.EditTransect(I);
  if Length(S) > 0 then
  begin
      TsectCombo.Text := S;
      TsectCombo.Items := Project.Lists[TRANSECT];
  end;
end;

procedure TXsectionForm.HelpBtnClick(Sender: TObject);
begin
  Application.HelpCommand(HELP_CONTEXT, 211450);
end;

end.
