unit Fgraph;

{-------------------------------------------------------------------}
{                    Unit:    Fgraph.pas                            }
{                    Project: EPA SWMM                              }
{                    Version: 5.0                                   }
{                    Date:    5/6/05     (5.0.005)                  }
{                             9/5/05     (5.0.006)                  }
{                             2/4/08     (5.0.012)                  }
{                             3/3/08     (5.0.013)                  }
{                    Author:  L. Rossman                            }
{                                                                   }
{   MDI child form that displays simulation results in the form     }
{   of a time series plot or a scatter plot. The graph is created   }
{   by calling CreateGraph. It gets refreshed (after a new analysis }
{   is made) by calling RefreshGraph. SetGraphOptions is called to  }
{   change display options, CopyTo to copy the graph to a file or   }
{   to the clipboard, and Print to print the graph.                 }
{                                                                   }
{   This unit uses several routines in the Ugraph unit.             }
{-------------------------------------------------------------------}

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, TeEngine, Series, ExtCtrls, TeeProcs, Chart, Math,
  Buttons, XPrinter, Uglobals, Uproject, Uutils;

type
  TGraphForm = class(TForm)
    Chart1: TChart;
    SpeedButton1: TSpeedButton;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure Chart1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
  private
    { Private declarations }
    StartPeriod:  LongInt;
    EndPeriod:    LongInt;
    TotalPeriods: LongInt;
    RescaleAxes:  Boolean;
    XvarIndex:    Integer;
    YvarIndex:    Integer;
    TimeFactor:   Double;
    StartDate:    TDateTime;
    EndDate:      TDateTime;
    OldStart:     TDateTime;
    OldEnd:       TDateTime;
    Graph:        TReportSelection;

////  These methods were moved to the Ucalib.pas unit. (LR - 9/5/05)  ////
//    function  AddCalibDataPoint(aSeries: TLineSeries; Tokens: TStringlist;
//              const DataTok: Integer): Boolean;
//    procedure AddCalibDataSeries(const Fname, ID: String; const K: Integer);

    function  CreateScatterPlot: Boolean;
    function  CreateTimeSeriesPlot: Boolean;
    procedure GetCalibSeries;
    procedure GetSeriesData(const K: Integer; Dt:TDateTime);
    procedure RefreshScatterPlot;
    procedure RefreshTimeSeriesPlot;
    procedure SetAxisScales;

////  New procedure added. (FM - 9/5/05) ////
    procedure FormatTimeAxis;

  public
    { Public declarations }
    procedure CopyTo;
    function  CreateGraph(ReportSelection: TReportSelection): Boolean;
    procedure Print(Destination: TDestination);
    procedure RefreshGraph;
    procedure SetGraphOptions;
  end;

//var
//  GraphForm: TGraphForm;

implementation

{$R *.DFM}

uses
  Fmain, Fmap, Ubrowser, Uoutput, Ugraph, Ucalib;

const
  TXT_GRAPH = 'Graph - ';
  TXT_SERIES = 'Series';
  TXT_COMPUTED = 'Computed';
  TXT_OBSERVED = 'Observed';
  TXT_TIME_HRS = 'Elapsed Time (hours)';
  TXT_TIME_DAYS = 'Elapsed Time (days)';
  TXT_SYS_RESULTS = 'System Results';
  TXT_SCATTER_LINE = 'Scatter Line';

  MSG_RETRIEVING_DATA = 'Retrieving Data Series ';
  MSG_RETRIEVING_CALIB_DATA = 'Retrieving calibration data... ';
  ObjectClasses: array[0..3] of String =
    ('Subcatchment', 'Node', 'Link', 'System');

//=============================================================================
//                        Event handlers
//=============================================================================

procedure TGraphForm.FormCreate(Sender: TObject);
//-----------------------------------------------------------------------------
// OnCreate handler for form.
//-----------------------------------------------------------------------------
begin
  // Create stringlists to store items being graphed
  // and names of data series being graphed
  Uglobals.SetFont(self);
  Graph.Items := TStringlist.Create;

  // Initialize chart properties
  Ugraph.InitGraphOptions(Chart1);
end;

procedure TGraphForm.FormClose(Sender: TObject; var Action: TCloseAction);
//-----------------------------------------------------------------------------
// OnClose handler for form.
//-----------------------------------------------------------------------------
begin
  Action := caFree;
end;

procedure TGraphForm.FormDestroy(Sender: TObject);
//-----------------------------------------------------------------------------
// OnDestroy handler for form. Frees chart series and stringlists.
//-----------------------------------------------------------------------------
begin
  with Chart1 do
    while SeriesCount > 0 do Series[0].Free;
  Graph.Items.Free;
end;

procedure TGraphForm.Chart1MouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
//-----------------------------------------------------------------------------
// OnMouseDown handler for Chart1. Launches graph options dialog
// when user right-clicks over the graph.
//-----------------------------------------------------------------------------
begin
  if (Button = mbRight) and not (ssCtrl in Shift)
  then SetGraphOptions;
end;

function  TGraphForm.CreateGraph(ReportSelection: TReportSelection): Boolean;
//-----------------------------------------------------------------------------
// Creates a graph based on selections in ReportSelection.
//-----------------------------------------------------------------------------
var
  I: Integer;
begin
  // Copy graph selection choices
  with ReportSelection do
  begin
    Graph.ReportType := ReportType;
    for I := 0 to VariableCount-1 do
      Graph.Variables[I] := Variables[I];
    Graph.ObjectType := ObjectType;
    Graph.XObjectType := XObjectType;
    Graph.StartDateIndex := StartDateIndex;
    Graph.EndDateIndex := EndDateIndex;
    Graph.DateTimeDisplay := DateTimeDisplay;
    if Assigned(Items) then Uutils.CopyStringList(Items, Graph.Items)
    else Graph.Items := nil;
    Graph.ItemCount := ItemCount;
    if ObjectType = SYS then Graph.ItemCount := VariableCount;
  end;

  // Determine number of time periods plotted
  StartPeriod := Uglobals.GetPeriod(Graph.StartDateIndex);
  EndPeriod := Uglobals.GetPeriod(Graph.EndDateIndex);
  TotalPeriods := EndPeriod - StartPeriod + 1;

  // Initialize old start/end dates
  RescaleAxes := True;
  OldStart := 0;
  OldEnd := 0;

  // Create the called for graph
  case Graph.ReportType of
    TIMESERIESPLOT: Result := CreateTimeSeriesPlot;
    SCATTERPLOT:    Result := CreateScatterPlot;
    else            Result := False;
  end;
end;

procedure TGraphForm.RefreshGraph;
//-----------------------------------------------------------------------------
// Refreshes graph display after data changes.
//-----------------------------------------------------------------------------
var
  K: Integer;
  Save_Cursor: TCursor;
begin
  // Exit if the graph is locked (SpeedButton1 is in the down position)
  if SpeedButton1.Down = True then exit;

  // If no results are available, then clear all previous data and
  // refresh the chart
  if not RunFlag then with Chart1 do
  begin
    for K := 0 to SeriesCount-1 do Series[K].Clear;
    Refresh;
    Exit;
  end;

  // Change mouse cursor shape
  Application.ProcessMessages;
  Save_Cursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;

  // Refresh the graph's data and redraw it
  try
    case Graph.ReportType of
      TIMESERIESPLOT: RefreshTimeSeriesPlot;
      SCATTERPLOT:    RefreshScatterPlot;
    end;
  finally
    Application.ProcessMessages;
    Screen.Cursor := Save_Cursor;  { Always restore to normal }
    MainForm.HideProgressBar;
  end;
end;


//=============================================================================
//                 Time Series Ploting Functions
//=============================================================================

function TGraphForm.CreateTimeSeriesPlot: Boolean;
//-----------------------------------------------------------------------------
// Creates a time series graph.
//-----------------------------------------------------------------------------
var
  K         : Integer;
  Title     : String;
  ObjName   : String;
  VarName   : String;
  VarUnits  : String;
  aSeries   : TChartSeries;

begin
  Result := True;
  with Graph do
  begin
    // Assign titles
    Uglobals.GetObjVarNames(ObjectType, Variables[0], ObjName, VarName, VarUnits);
    if Length(VarUnits) > 0 then VarUnits := ' (' + VarUnits + ')';
    if ObjectType = SYS then
    begin
      if ItemCount = 1 then Title := ObjName + ' ' + VarName
      else Title := TXT_SYS_RESULTS;
    end
    else
    begin
      if ItemCount = 1 then Title := ObjName + ' ' + Items[0] + ' ' + VarName
      else Title := ObjectClasses[ObjectType] + ' ' + VarName;
    end;
    Caption := TXT_GRAPH + Title;
    Chart1.Title.Text.Add(Title);
    if (ObjectType = SYS) and (ItemCount > 1) then
      Chart1.LeftAxis.Title.Caption := ''
    else
      Chart1.LeftAxis.Title.Caption := VarName + VarUnits;
    if ItemCount > 1 then Chart1.Legend.Visible := True;

    // Consider each object being plotted
    Title := TXT_COMPUTED;
    for K := 0 to ItemCount - 1 do
    begin

      // Get the names of the objects being plotted
      if ItemCount > 1 then
      begin
        if ObjectType = SYS then
        begin
          Uglobals.GetObjVarNames(ObjectType, Variables[K], ObjName, VarName, VarUnits);
          Title := VarName + ' - ' + VarUnits;
        end
        else Title := ObjName + ' ' + Items[K];
      end;

      // Create a Fast Line data series for large time series or a normal
      // Line series otherwise
      if TotalPeriods >= 32000
      then aSeries := Ugraph.CreateFastLineSeries(Chart1, Title)
      else aSeries := Ugraph.CreateLineSeries(Chart1, Title);
      if Assigned(aSeries) then aSeries.Tag := K;

      // For speed, do not order the X-values
      Chart1.Series[K].XValues.Order := loNone;

      // Use a DateTime X-variable if so specified
      Chart1.Series[K].XValues.DateTime := DateTimeDisplay;
    end;
    Chart1.Legend.Visible := (ItemCount > 1);                                   //(5.0.013 - LR)

    // Save index of variable being plotted
    YvarIndex := Uoutput.GetVarIndex(Variables[0], ObjectType);

    // Create a line series for calibration data
    if ItemCount = 1 then
    begin
      aSeries := Ugraph.CreateLineSeries(Chart1, TXT_OBSERVED);
      if Assigned(aSeries) then
      begin
        aSeries.Tag := MAXSERIES;
        aSeries.ShowInLegend := False;
        Ugraph.InitLineSeriesOptions(TLineSeries(aSeries), MAXSERIES);
      end;
    end;

    // Display the empty plot
//    Refresh;                                                                  //(5.0.013 - LR)
  end;
end;

////  New procedure added to format time axis. (FM - 9/5/05)  ////
procedure TGraphForm.FormatTimeAxis;
//-----------------------------------------------------------------------------
// Formats the time axis of a time series plot.
//-----------------------------------------------------------------------------
var
  DeltaDays : Integer;
begin
  with Graph do
  begin
    // Determine Time units
    DeltaDays := Round(EndDate - StartDate);
    TimeFactor := 1.0;

    if not DateTimeDisplay then
    begin
      if DeltaDays <= 7 then
      begin
        Chart1.BottomAxis.Title.Caption := TXT_TIME_HRS;
        TimeFactor := 24.0;
      end
      else Chart1.BottomAxis.Title.Caption := TXT_TIME_DAYS;
    end
    else begin
      Chart1.BottomAxis.Title.Caption := ' ';
      if Length(GraphOptions.DateTimeFormat) > 0 then                           //(5.0.012 - LR)
        Chart1.BottomAxis.DateTimeFormat := GraphOptions.DateTimeFormat         //(5.0.012 - LR)
      else if DeltaDays <= 3 then                                               //(5.0.012 - LR)
        Chart1.BottomAxis.DateTimeFormat := 'h:nn'                              //(5.0.012 - LR)
      else if DeltaDays <= 10 then                                              //(5.0.012 - LR)
        Chart1.BottomAxis.DateTimeFormat := 'h:nn m/d/yy'
      else if DeltaDays <= 60 then
        Chart1.BottomAxis.DateTimeFormat := 'm/d yyyy'
      else if DeltaDays <= 10*365 then
        Chart1.BottomAxis.DateTimeFormat := 'mmm yyyy'
      else
         Chart1.BottomAxis.DateTimeFormat := 'yyyy';
    end;
  end;

end;


procedure TGraphForm.RefreshTimeSeriesPlot;
//-----------------------------------------------------------------------------
// Refreshes drawing of a time series plot.
//-----------------------------------------------------------------------------
var
  K          : Integer;
  HasPoints  : Boolean;
  Dt         : TDateTime;
begin

  // Determine number of time periods plotted
  StartPeriod := Uglobals.GetPeriod(Graph.StartDateIndex);
  EndPeriod := Uglobals.GetPeriod(Graph.EndDateIndex);
  TotalPeriods := EndPeriod - StartPeriod + 1;

  // Determine range of dates to plot
  Dt := DeltaDateTime;
  StartDate := Floor(StartDateTime + StartPeriod*Dt);
  EndDate := Ceil(StartDateTime + EndPeriod*Dt);
  HasPoints := False;

////  New procedure used to format time axis. (FM - 9/5/05)  ////
  FormatTimeAxis;

  // Inactivate each data series (so that graph doesn't keep redrawing itself)
  for K := 0 to Chart1.SeriesCount-1 do
  begin
    Chart1.Series[K].Active := False;
    Chart1.Series[K].Clear;
  end;

  // Set the TChart variable that sizes the series being plot
  TeeDefaultCapacity := TotalPeriods;

  // Retrieve data for each time series in the graph
  for K := 0 to Graph.ItemCount - 1 do
  begin
    MainForm.ShowProgressBar(MSG_RETRIEVING_DATA + IntToStr(K+1) + '... ');
    GetSeriesData(K, Dt);
    if Chart1.Series[K].Count > 0 then HasPoints := True;
  end;

  // Re-set the chart series' default size
  TeeDefaultCapacity := 0;

  // Add calibration data for a single-object plot
  if (Graph.ItemCount = 1) then GetCalibSeries;

  // Adjust min & max values of axes
  if RescaleAxes and HasPoints then SetAxisScales
  else with Chart1 do
  begin
    if (OldStart <> StartDate)
    or (OldEnd <> EndDate)
    or (LeftAxis.Minimum > MinYValue(LeftAxis))
    or (LeftAxis.Maximum < MaxYValue(LeftAxis))
    then SetAxisScales;
  end;

  // Activate each series
  for K := 0 to Chart1.SeriesCount-1 do
  begin
    if Chart1.Series[K].Count > 0 then Chart1.Series[K].Active := True;
  end;
  OldStart := StartDate;
  OldEnd := EndDate;
end;

procedure TGraphForm.GetSeriesData(const K: Integer; Dt: TDateTime);
//-----------------------------------------------------------------------------
//  Loads time series data into chart series where  K = chart series index
//  and Dt = time increment of output values in decimal days.
//-----------------------------------------------------------------------------
var
  J          : LongInt;
  theObject  : TObject;
  X, Y       : Double;
  Progress   : LongInt;
  ProgStep   : LongInt;

begin
  ProgStep := TotalPeriods div (MainForm.ProgressBar.Max div
                                MainForm.ProgressBar.Step);
  Progress := 0;
  with Graph do
  begin
    // Get a reference to the object being plotted
    if ObjectType <> SYS then
    begin
      theObject := Uoutput.GetObject(ObjectType, Items[K]);
      if theObject = nil then Exit;
    end
    else theObject := nil;

    // Retrieve values to be plotted from the simulation's binary output file
    try
      // Find starting X value for graph
      with Chart1.Series[K] do
      begin

        // Find initial date of graph
        if DateTimeDisplay then X := StartDateTime + StartPeriod*Dt

        // Find initial elapsed time
        else
        begin
          Dt := Dt*TimeFactor;
          X := (StartPeriod+1)*Dt;
        end;

        // Determine the index of the plotted variable
        // (in the binary output file)
        if ObjectType = SYS
        then YvarIndex := Uoutput.GetVarIndex(Variables[K], ObjectType);

        // Get Y-value for each reporting period in graph
        for J := StartPeriod to EndPeriod do
        begin
          // If a Y-value found, add it to the graph
          Y := Uoutput.GetValue(ObjectType, YvarIndex, J, theObject);
          if Y <> MISSING then AddXY(X, Y, '', clTeeColor);

          // Move to next time X
          X := X + Dt;
          MainForm.UpdateProgressBar(Progress, ProgStep);
        end;

      end;
    finally
    end;
  end;
end;

////  This procedure has been completely rewritten. (LR - 9/5/05)  ////
procedure TGraphForm.GetCalibSeries;
//-----------------------------------------------------------------------------
//  Loads calibration data into a chart series.
//-----------------------------------------------------------------------------
var
  Nseries  : Integer;
begin
  // Retrieve data points from the calibration file
  MainForm.ShowProgressBar(MSG_RETRIEVING_CALIB_DATA);
  Nseries := Chart1.SeriesCount-1;
  Chart1.Series[Nseries].XValues.Order := loNone;
  if Graph.Items.Count = 0 then exit;
  Ucalib.GetCalibData(Graph.Variables[0], Graph.ObjectType, Graph.Items[0],
    StartDate, EndDate, TimeFactor, Graph.DateTimeDisplay,
    Chart1.Series[Nseries]);

  // Activate the chart's calibration data series legend.
  with Chart1.Series[Nseries] do
  begin
    if (Count > 0) then
    begin
      ShowInLegend := True;
      Chart1.Legend.Visible := True;
    end;
  end;
end;

//=============================================================================
//                     Scatter Plot Functions
//=============================================================================

function TGraphForm.CreateScatterPlot: Boolean;
//-----------------------------------------------------------------------------
// Creates a scatter plot graph.
//-----------------------------------------------------------------------------
var
  XObjName  : String;
  YObjName  : String;
  XVarName  : String;
  YVarName  : String;
  XVarUnits : String;
  YVarUnits : String;
  Title     : String;
begin
  Result := True;
  with Graph do
  begin
    // Assign title strings
    XvarIndex := Uoutput.GetVarIndex(Variables[0], XObjectType);
    YvarIndex := Uoutput.GetVarIndex(Variables[1], ObjectType);
    Uglobals.GetObjVarNames(XObjectType, Variables[0], XObjName, XVarName, XVarUnits);
    Uglobals.GetObjVarNames(ObjectType, Variables[1], YObjName, YVarName, YVarUnits);
    if Length(XVarUnits) > 0 then XVarUnits := ' (' + XVarUnits + ')';
    if Length(YVarUnits) > 0 then YVarUnits := ' (' + YVarUnits + ')';
    Title := YObjName + ' ' + Items[1] + ' ' + YVarName + ' v. ' +
             XObjName + ' ' + Items[0] + ' ' + XVarName;
    Caption := 'Graph - ' + Title;
    Chart1.Title.Text.Add(Title);
    Chart1.BottomAxis.Title.Caption := XObjName + ' ' + Items[0] + ' ' +
                                       XVarName + XVarUnits;
    Chart1.LeftAxis.Title.Caption := YObjName + ' ' + Items[1] + ' ' +
                                     YVarName + YVarUnits;
  end;

  // Create a Line series for the plot & set its options
  Ugraph.CreateLineSeries(Chart1, TXT_SCATTER_LINE);
  with Chart1 do
  begin
    Legend.Visible := False;
    with Series[0] as TLineSeries do
    begin
      XValues.Order := loNone;
    end;
  end;

  // Display the empty plot
//  Refresh;                                                                    //(5.0.013 - LR)
end;

procedure TGraphForm.RefreshScatterPlot;
//-----------------------------------------------------------------------------
// Refreshes the drawing of a scatter plot.
//-----------------------------------------------------------------------------
var
  J          : LongInt;
  Xobject    : TObject;
  Yobject    : TObject;
  X, Y       : Double;
  HasPoints  : Boolean;
  Progress   : LongInt;
  ProgStep   : LongInt;
begin
  // Determine number of time periods plotted
  StartPeriod := Uglobals.GetPeriod(Graph.StartDateIndex);
  EndPeriod := Uglobals.GetPeriod(Graph.EndDateIndex);
  TotalPeriods := EndPeriod - StartPeriod + 1;
  HasPoints := False;
  ProgStep := TotalPeriods div (MainForm.ProgressBar.Max div
                                MainForm.ProgressBar.Step);
  Progress := 0;
  MainForm.ShowProgressBar('Retrieving data... ');

  // Set the default series size for a TChart's data series
  TeeDefaultCapacity := TotalPeriods;

  // Add each pair of X & Y values to the series being plotted
  with Graph do
  begin
    // Get references to the X-object & Y-Object being plotted
    Xobject := Uoutput.GetObject(XObjectType, Items[0]);
    Yobject := Uoutput.GetObject(ObjectType, Items[1]);
    if (Xobject = nil) or (Yobject = nil) then Exit;

     // Add data to the chart series
    with Chart1.Series[0] do
    begin
      Clear;
      Active := False;
      TeeDefaultCapacity := TotalPeriods;
      for J := StartPeriod to EndPeriod do
      begin
        X := Uoutput.GetValue(XObjectType, XvarIndex, J, Xobject);
        Y := Uoutput.GetValue(ObjectType, YvarIndex, J, Yobject);
        if (X <> MISSING) and (Y <> MISSING) then AddXY(X, Y, '', clTeeColor);
        MainForm.UpdateProgressBar(Progress, ProgStep);
      end;
      if Count > 0 then HasPoints := True;
    end;
  end;

  // Reset the default size of a TChart's series
  TeeDefaultCapacity := 0;

  // Adjust min & max values of axes
  if RescaleAxes and HasPoints then SetAxisScales
  else with Chart1 do
  begin
    if (LeftAxis.Minimum > MinYValue(LeftAxis))
    or (LeftAxis.Maximum < MaxYValue(LeftAxis))
    or (BottomAxis.Minimum > MinXValue(BottomAxis))
    or (BottomAxis.Maximum < MaxXValue(BottomAxis))
    then SetAxisScales;
  end;

  with Chart1.Series[0] as TLineSeries do
  begin
    LinePen.Visible := False;
    Pointer.Visible := True;
  end;
  Chart1.Series[0].Active := True;
end;

procedure TGraphForm.SetAxisScales;
//-----------------------------------------------------------------------------
// Sets X and Y axis scaling.
//-----------------------------------------------------------------------------
begin
  Ugraph.SetAxisScale(Chart1.LeftAxis, 0, 0, 0);
  if not Graph.DateTimeDisplay
  then Ugraph.SetAxisScale(Chart1.BottomAxis, 0, 0, 0)
  else Chart1.BottomAxis.Automatic := True;
  RescaleAxes := False;
end;

////  This procedure has been updated.  ////                                    //(5.0.012 - LR)
procedure TGraphForm.SetGraphOptions;
//-----------------------------------------------------------------------------
// Launches the Graph Options dialog form.
//-----------------------------------------------------------------------------
var
  OldDateTimeFormat: String;
begin
  OldDateTimeFormat := GraphOptions.DateTimeFormat;
  Ugraph.SetGraphOptions(Chart1, MainForm.ChartDialog);
  if not SameText(OldDateTimeFormat, GraphOptions.DateTimeFormat)
  then Uglobals.HasChanged := True;
end;

procedure TGraphForm.CopyTo;
//-----------------------------------------------------------------------------
// Launches a CopyTo dialog form to copy the graph to a file or to the
// clipboard.
//-----------------------------------------------------------------------------
begin
  Ugraph.CopyTo(Chart1);
end;

procedure TGraphForm.Print(Destination: TDestination);
//-----------------------------------------------------------------------------
// Prints the graph to the printer or to a preview window.
//-----------------------------------------------------------------------------
begin
  Ugraph.Print(Chart1, MainForm.thePrinter, Destination);
end;

end.
