unit Fstatus;

{-------------------------------------------------------------------}
{                    Unit:    Fstatus.pas                           }
{                    Project: EPA SWMM                              }
{                    Version: 5.0                                   }
{                    Date:    7/5/06      (5.0.008)                 }
{                             2/4/08      (5.0.012)                 }
{                             1/21/09     (5.0.014)                 }
{                    Author:  L. Rossman                            }
{                                                                   }
{   MDI child form that displays the status report generated        }
{   from a run of SWMM.                                             }
{                                                                   }
{   This unit was totally rewritten to make use of the Orpheus      }
{   FileViewer control. (LR - 7/5/06 and 5.0.014)                   }
{                                                                   }
{   Changes for 5.0.012:                                            }
{   A listbox with topic headings was added to make it easier to    }
{   navigate to a specific topic in the status report display.      }
{-------------------------------------------------------------------}

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, ComCtrls, Clipbrd, Menus, StrUtils, Xprinter,
  Uglobals, Uutils,
  ovcbase, ovcviewr, ovcdata, ExtCtrls;              //Orpheus units

type
  TStatusForm = class(TForm)
    FindDialog1: TFindDialog;
    PageControl1: TPageControl;
    TabSheet1: TTabSheet;
    ListBox1: TListBox;
    Splitter1: TSplitter;
    FileViewer: TOvcFileViewer;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FindDialog1Find(Sender: TObject);
    procedure ListBox1Click(Sender: TObject);
  private
    { Private declarations }
    procedure LoadTopics;                                                       //(5.0.012 - LR)
  public
    { Public declarations }
    procedure RefreshStatusReport;
    procedure SetFont;
    procedure FindObject;
    procedure FindText;
    procedure SelectAll;
    procedure CopyTo;
    procedure Print(Destination: TDestination);
  end;

var
  StatusForm: TStatusForm;        // Do not comment out this line.

implementation

{$R *.DFM}

uses
  Dcopy, Dfind, Fmain;

const
  MSG_NO_FILE = 'There is no Report File to view.';                             //(5.0.012 - LR)
  MSG_REPORT_TOO_BIG = 'Status Report too big to fit in window.';

//// These arrays were added to support the topic listbox. ////                 //(5.0.012 - LR)
  Topics: array[0..19] of String =                                              //(5.0.014 - LR)
    ('Analysis Options',    'Input Summary',       'Rain File Summary',
     'Warning Messages',                                                        //(5.0.014 - LR)
     'Error Messages',      'Control Actions',     'Continuity Errors',
     'Stability Results',   'Runoff Results',      'Washoff Results',
     'Node Depths',         'Node Inflows',        'Node Surcharging',
     'Node Flooding',       'Storage Volumes',     'Outfall Loadings',
     'Link Flows',          'Flow Classification', 'Conduit Surcharging',
     'Pumping Summary');

  TopicHdrs: array[0..19] of String =                                           //(5.0.014 - LR)
    ('Analysis Options',               'Element Count',
     'Rainfall File Summary',          '  WARNING',          '  ERROR',         //(5.0.014 - LR)
     'Control Actions',                'Continuity',
     'Highest Flow Instability',       'Subcatchment Runoff Summary',
     'Subcatchment Washoff Summary',   'Node Depth Summary',
     'Node Inflow Summary',            'Node Surcharge Summary',
     'Node Flooding Summary',          'Storage Volume Summary',
     'Outfall Loading Summary',        'Link Flow Summary',
     'Flow Classification Summary',    'Conduit Surcharge Summary',
     'Pumping Summary');

var
  TopicPos: array[0..19] of LongInt;                                            //(5.0.012 - LR)

procedure TStatusForm.FormClose(Sender: TObject; var Action: TCloseAction);
//-----------------------------------------------------------------------------
// OnClose handler for form.
//-----------------------------------------------------------------------------
begin
  Action := caFree;
end;

procedure TStatusForm.ListBox1Click(Sender: TObject);                           //(5.0.012 - LR)
//-----------------------------------------------------------------------------
// OnClick handler for the topics listbox. Positions the status report
// file viewer to the line where the selected topic begins.
//-----------------------------------------------------------------------------
var
  I: Integer;
begin
  I := ListBox1.ItemIndex;
  if TopicPos[I] >= 0 then FileViewer.TopLine := TopicPos[I];
end;

procedure TStatusForm.RefreshStatusReport;
//-----------------------------------------------------------------------------
// Reloads the FileViewer control with the contents
// of the Status Report File generated from an analysis.
//-----------------------------------------------------------------------------
begin
  // Set the FileViewer's font & clear its contents
  FileViewer.IsOpen := False;
  SetFont;

  // Make sure that the report file exists
  if not FileExists(TempReportFile)
  then PageControl1.Visible := False
  else with FileViewer do
  try
    // Load the contents of the file into the FileViewer
    FileName := TempReportFile;
    IsOpen := True;
    LoadTopics;                                                                 //(5.0.012 - LR)
  except
    On E: Exception do
      MessageDlg(MSG_NO_FILE, mtWarning, [mbOK], 0);                            //(5.0.012 - LR)
  end;
end;

procedure TStatusForm.SetFont;
//-----------------------------------------------------------------------------
// Sets font style of the FileViewer control.
//-----------------------------------------------------------------------------
begin
  Uglobals.SetFont(self);
  //Fileviewer.FixedFont.Size := Font.Size;
  Fileviewer.FixedFont.Style := Font.Style;
end;

procedure TStatusForm.FindObject;
//-----------------------------------------------------------------------------
// Launches the Map Finder form to locate an object whose ID name
// has been selected in the FileViewer.
//-----------------------------------------------------------------------------
var
  N: Integer;
  S: String;
begin
  FileViewer.CopyToClipboard;
  if Clipboard.HasFormat(CF_TEXT) then
  begin
    S := TrimLeft(Clipboard.AsText);
    N := Pos(' ', S);
    if N > 0 then S := LeftStr(S, N-1);
    FindForm.Visible := True;
    FindForm.Edit1.Text := S;
  end;
end;

procedure TStatusForm.SelectAll;
//-----------------------------------------------------------------------------
// Selects all text in the FileViewer.
//-----------------------------------------------------------------------------
begin
  FileViewer.SelectAll(False);
end;

procedure TStatusForm.FindDialog1Find(Sender: TObject);
//-----------------------------------------------------------------------------
// OnFind handler for the Find Text dialog.
//-----------------------------------------------------------------------------
var
  OptionSet: TSearchOptionSet;  //Defined in the Orpheus OvcData unit.
begin
  OptionSet := [];
  OptionSet := OptionSet + [soFind] + [soSelText];
  with FindDialog1 do
  begin
    if frWholeWord in Options then OptionSet := OptionSet + [soWholeWord];
    if frMatchCase in Options then OptionSet := OptionSet + [soMatchCase];
    if not (frDown in Options) then OptionSet := OptionSet + [soBackward];
    if not FileViewer.Search(FindText, OptionSet) then
      MessageDlg('Search string ''' + FindText + ''' not found',
        mtInformation, [mbOk], 0);
  end;
end;

////  This is a new procedure added to release 5.0.012.  ////                   //(5.0.012 - LR)
procedure TStatusForm.LoadTopics;
//-----------------------------------------------------------------------------
// Saves the line position where each major topic of the Status Report begins.
//-----------------------------------------------------------------------------
var
  I: Integer;
  N: Integer;
  C: Integer;
begin
  FileViewer.Visible := False;
  ListBox1.Clear;
  N := 0;
  for I := 0 to High(TopicHdrs) do TopicPos[I] := -1;
  for I := 0 to High(TopicHdrs) do
  begin
    if FileViewer.Search(TopicHdrs[I], [soFind]-[soSelText]) then
    begin
      FileViewer.SetMarker(0);
      TopicPos[N] := FileViewer.GetMarkerPosition(0, C) - 1;
      ListBox1.Items.Add(Topics[I]);
      Inc(N);
    end
    else FileViewer.TopLine := 0;
  end;
  if N > 0 then
  begin
    FileViewer.ClearMarker(0);
    FileViewer.SetSelection(0,0,0,0,False);
    PageControl1.Visible := True;
    ListBox1.ItemIndex := 0;
    ListBox1.SetFocus;
  end
  else PageControl1.Visible := False;
  FileViewer.TopLine := 0;
  FileViewer.Visible := True;
end;

procedure TStatusForm.FindText;
//-----------------------------------------------------------------------------
// Executes a Find Text Dialog to locate selected text in the FileViewer.
//-----------------------------------------------------------------------------
var
  L, T: Integer;
begin
  L := MainForm.Left + (MainForm.Width - MainForm.ClientWidth) - 2;
  T := MainForm.Top + (MainForm.Height-MainForm.ClientHeight) - 2;
  with FindDialog1 do
  begin
    Position := Point(L,T);
    Execute;
  end;
end;

procedure TStatusForm.CopyTo;
//-----------------------------------------------------------------------------
// Copies contents of the FileViewer to either a file or to the Clipboard.
//-----------------------------------------------------------------------------
var
  L1, L2: LongInt;
  C1, C2: Integer;
begin
  // Create the CopyTo dialog form
  with TCopyToForm.Create(self) do
  try

    // Disable format selection (since it has to be Text)
    FormatGroup.ItemIndex := 2;
    FormatGroup.Enabled := False;

    // Show the form modally
    if ShowModal = mrOK then with FileViewer do
    begin

      // If user supplies a file name then copy contents of FileViewer to it
      if Length(DestFileName) > 0
      then CopyFile(PChar(TempReportFile), PChar(DestFileName), FALSE)

      // Otherwise copy the contents into the Clipboard
      else
      begin
        if not FileViewer.GetSelection(L1, C1, L2, C2)
        then FileViewer.SelectAll(False);
        FileViewer.CopyToClipboard;
      end;
    end;

  // Free the CopyTo dialog form
  finally
    Free;
  end;
end;

procedure TStatusForm.Print(Destination: TDestination);
//-----------------------------------------------------------------------------
// Prints Status Report to Destination (printer or preview form).
//-----------------------------------------------------------------------------
var
  I    : Integer;
  S    : array[0..1024] of AnsiChar;
begin
  with MainForm.thePrinter do
  begin
    BeginJob;
    SetDestination(Destination);
    with FileViewer.FixedFont do
      SetFontInformation(Name, Size, Style);
    with FileViewer do
      for I := 0 to LineCount - 1 do
      begin
        GetPrintableLine(I, S, SizeOf(S));
        PrintLine(S);
      end;
    EndJob;
  end;
end;

end.
